'use strict';

const { expect } = require('chai');
const sinon = require('sinon');
const rewire = require('rewire');
const { getPackageName } = require('exframe-utilities');
const util = require('./support/util');
const { loggerMock } = require('./support/loggerMock');

context('Service', () => {
  describe('Unit Tests', () => {
    beforeEach(() => {
      this.settings = {
        logger: loggerMock,
        timeout: 10
      };

      this.sinon = sinon.createSandbox();
      this.clock = sinon.useFakeTimers();
      this.service = rewire('../index');
      this.health = require('exframe-health');
    }, 10000);

    afterEach(() => {
      this.health.healthChecks = {};
      this.sinon.restore();
      this.clock.restore();
    });

    describe('init', () => {
      it('init test', () => {
        const initSpy = this.sinon.spy(this.service, 'init');
        initSpy(this.settings);
        expect(initSpy.calledOnce).to.be.true;
      });

      it('passes config object to the memory watcher module\'s "getInstance" function', async () => {
        const testOptions = {
          logger: { info: this.sinon.stub() },
          memoryWatcher: { test: true }
        };
        const getInstanceMock = this.sinon.stub().returns({
          start: () => {}
        });
        this.service.__set__('MemoryWatcher', {
          getInstance: getInstanceMock
        });
        this.service.init(testOptions);
        expect(getInstanceMock.callCount).to.equal(1);
        expect(getInstanceMock.getCall(0).args[0]).to.deep.equal({
          ...testOptions.memoryWatcher,
          logger: testOptions.logger
        });
      });

      it('starts the memory watcher', async () => {
        const startMock = this.sinon.stub();
        this.service.__set__('MemoryWatcher', {
          getInstance: this.sinon.stub().returns({
            start: startMock
          })
        });
        this.service.init();
        expect(startMock.callCount).to.equal(1);
      });

      it('adds a health check', async () => {
        this.service.__set__('MemoryWatcher', {
          getInstance: this.sinon.stub().returns({
            start: this.sinon.stub()
          })
        });
        const healthSpy = this.sinon.spy(this.health, 'add');

        this.service.init();
        expect(healthSpy.callCount).to.equal(1);
        expect(healthSpy.getCall(0).args[0]).to.equal('memory-watcher');
      });
    });

    it('settings test', () => {
      const initSpy = this.sinon.spy(this.service, 'init');
      initSpy(this.settings);
      expect(this.service.settings.timeout).to.eql(this.settings.timeout);
      expect(this.service.settings.logger).to.deep.eql(this.settings.logger);
    });

    it('status test', () => {
      expect(this.service.STATUS.RUNNING).to.be.eql('Running');
      expect(this.service.STATUS.SHUTTING_DOWN).to.be.eql('ShuttingDown');
    });

    it('process.on test', () => {
      this.sinon.stub(this.service.process, 'on');
      process.on.callsFake(() => true);
      expect(process.on()).to.eql(true);
    });

    it('onSigterm test', () => {
      this.sinon.stub(this.service.process, 'exit');
      const initSpy = this.sinon.spy(this.service, 'init');
      const gracefulShutdownStub = this.sinon.stub();

      this.service.__set__({
        gracefulShutdown: gracefulShutdownStub
      });

      gracefulShutdownStub.returns(true);

      initSpy(this.settings);
      this.service.onSigterm();
      expect(gracefulShutdownStub.calledOnce).to.be.false;
      this.clock.tick(10);
      expect(gracefulShutdownStub.calledOnce).to.be.true;
    });

    it('onSigterm test, gracefulShutdown is called once', () => {
      this.sinon.stub(this.service.process, 'exit');
      const initSpy = this.sinon.spy(this.service, 'init');
      const gracefulShutdownStub = this.sinon.stub();

      this.service.__set__({
        gracefulShutdown: gracefulShutdownStub
      });

      gracefulShutdownStub.returns(true);

      initSpy(this.settings);
      this.service.onSigterm();
      expect(gracefulShutdownStub.calledOnce).to.be.false;
      this.clock.tick(10);
      expect(gracefulShutdownStub.calledOnce).to.be.true;
      this.service.onSigterm();
      this.clock.tick(10);
      expect(gracefulShutdownStub.calledOnce).to.be.true;
    });

    it('gracefulShutdown test', async () => {
      const mqSignal = () => new Promise((resolve) => {
        setTimeout(
          resolve(true),
          50000
        );
      });

      const restSignal = () => new Promise((resolve) => {
        resolve(true);
      });

      const dbSignal = () => new Promise((resolve) => {
        setTimeout(
          resolve({ message: 'DB closed' }),
          10000
        );
      });

      const resourceStatus1 = {
        name: 'db',
        status: this.service.STATUS.RUNNING,
        signal: dbSignal,
        order: 'last'
      };

      const resourceStatus2 = {
        name: 'mq',
        status: this.service.STATUS.RUNNING,
        signal: mqSignal,
        order: 'general'
      };

      const resourceStatus3 = {
        name: 'rest',
        status: this.service.STATUS.RUNNING,
        signal: restSignal,
        order: 'general'
      };

      const initSpy = this.sinon.spy(this.service, 'init');
      initSpy(this.settings);
      this.service.serviceStatus.STATUS = this.service.STATUS.RUNNING;
      this.service.serviceStatus.resources.push(resourceStatus1, resourceStatus2, resourceStatus3);
      const processExitSpy = this.sinon.stub(this.service.process, 'exit');
      await this.service.gracefulShutdown();
      expect(processExitSpy.calledOnce).to.be.true;
    });

    it('gracefulShutdown failure test', async () => {
      const aSignal = () => new Promise((resolve, reject) => {
        reject(new Error());
      });

      const resourceStatus = {
        name: 'mq',
        status: this.service.STATUS.RUNNING,
        signal: aSignal,
        order: 'general'
      };

      const initSpy = this.sinon.spy(this.service, 'init');
      initSpy(this.settings);
      this.service.serviceStatus.STATUS = this.service.STATUS.RUNNING;
      this.service.serviceStatus.resources.push(resourceStatus);
      const processExitSpy = this.sinon.stub(this.service.process, 'exit');
      await this.service.gracefulShutdown();
      expect(processExitSpy.calledOnce).to.be.false;
    });

    it('runOnSignal test, with three registered services', async () => {
      const mqSignal = () => new Promise((resolve) => {
        setTimeout(
          resolve(true),
          200
        );
      });

      const restSignal = () => new Promise((resolve) => {
        resolve(true);
      });

      const dbSignal = () => new Promise((resolve) => {
        setTimeout(
          resolve({ message: 'DB closed' }),
          100
        );
      });

      const resultArray = [
        true, // built-in memory watcher resource
        true,
        true,
        { message: 'DB closed' }
      ];

      const resourceStatus1 = {
        name: 'db',
        status: this.service.STATUS.RUNNING,
        signal: dbSignal,
        order: 'last'
      };

      const resourceStatus2 = {
        name: 'mq',
        status: this.service.STATUS.RUNNING,
        signal: mqSignal,
        order: 'general'
      };

      const resourceStatus3 = {
        name: 'rest',
        status: this.service.STATUS.RUNNING,
        signal: restSignal,
        order: 'general'
      };

      const initSpy = this.sinon.spy(this.service, 'init');
      initSpy(this.settings);
      this.service.serviceStatus.STATUS = this.service.STATUS.RUNNING;
      this.service.serviceStatus.resources.push(resourceStatus1, resourceStatus2, resourceStatus3);
      this.sinon.stub(this.service.process, 'exit');
      const sortedResources = await this.service.sortResources();
      const result = await this.service.runOnSignal();
      for (let i = 0; i < sortedResources.length; i++) {
        expect(resultArray[i]).to.eql(result[i]);
      }
    });

    it('runOnSignal test, return two services with same order type', async () => {
      const mqSignal = () => new Promise((resolve) => {
        setTimeout(
          resolve(true),
          100
        );
      });

      const restSignal = () => new Promise((resolve) => {
        resolve(false);
      });

      const resultArray = [
        true, // built-in memory watcher resource
        true,
        false
      ];

      const resourceStatus1 = {
        name: 'mq',
        status: this.service.STATUS.RUNNING,
        signal: mqSignal,
        order: 'general'
      };

      const resourceStatus2 = {
        name: 'rest',
        status: this.service.STATUS.RUNNING,
        signal: restSignal,
        order: 'general'
      };

      const initSpy = this.sinon.spy(this.service, 'init');
      initSpy(this.settings);
      this.service.serviceStatus.STATUS = this.service.STATUS.RUNNING;
      this.service.serviceStatus.resources.push(resourceStatus1, resourceStatus2);
      this.sinon.stub(this.service.process, 'exit');
      const sortedResources = await this.service.sortResources();
      const result = await this.service.runOnSignal();
      for (let i = 0; i < sortedResources.length; i++) {
        expect(resultArray[i]).to.eql(result[i]);
      }
    });

    it('runOnSignal test, no resources registered', async () => {
      const initSpy = this.sinon.spy(this.service, 'init');
      initSpy(this.settings);
      this.service.serviceStatus.STATUS = this.service.STATUS.RUNNING;
      this.service.serviceStatus.resources.push();
      this.sinon.stub(this.service.process, 'exit');
      await this.service.sortResources();
      const result = await this.service.runOnSignal();
      expect(result).to.have.lengthOf(1); // built-in memory watcher resource
    });

    it('exports a prom-client object to use for adding prometheus metrics', () => {
      expect(this.service.prometheusClient).to.have.property('register');
    });

    it('enables default process metrics', async () => {
      const metrics = await this.service.prometheusClient.register.getMetricsAsJSON();
      const testMetric = metrics.find((metric) => metric.name === 'process_cpu_seconds_total');
      expect(testMetric).to.exist;
    });

    it('sets the "serviceName" label to the npm_package_name env var for metrics in the default register', async () => {
      const c = new this.service.prometheusClient.Counter({ // eslint-disable-line no-unused-vars
        name: 'test_counter',
        help: 'test counter'
      });

      const metrics = await this.service.prometheusClient.register.getMetricsAsJSON();
      const testMetric = metrics.find((metric) => metric.name === 'test_counter');
      expect(testMetric.values[0].labels.serviceName).to.equal(getPackageName());
    });

    describe('default health check', () => {
      it('reports healthy when the "MemoryWatcher.liveness" property returns "true"', async () => {
        this.service.__set__('MemoryWatcher', {
          getInstance: this.sinon.stub().returns({
            liveness: true,
            start: () => {}
          })
        });
        const res = { status: this.sinon.stub(), json: this.sinon.stub() };
        this.service.init();

        await this.health.check({}, res);
        expect(res.json.callCount).to.equal(1);
        expect(res.json.firstCall.args[0]).to.deep.equal({
          version: 'unknown',
          dependencies: {
            service: { status: 200, message: 'OK' },
            'memory-watcher': { status: 200, message: 'OK' }
          },
          state: 'Healthy'
        });
      });

      it('reports unhealthy when the "MemoryWatcher.liveness" property returns "false"', async () => {
        this.service.__set__('MemoryWatcher', {
          getInstance: this.sinon.stub().returns({
            liveness: false,
            start: () => {}
          })
        });
        const res = { status: this.sinon.stub(), json: this.sinon.stub() };
        this.service.init();

        await this.health.check({}, res);
        expect(res.json.callCount).to.equal(1);
        expect(res.json.firstCall.args[0]).to.deep.equal({
          version: 'unknown',
          dependencies: {
            service: { status: 200, message: 'OK' },
            'memory-watcher': { status: 503, message: 'Liveness memory limit exceeded' }
          },
          state: 'Unhealthy'
        });
      });
    });

    describe('liveness check', () => {
      it('reports healthy when the "MemoryWatcher.liveness" property returns "true"', async () => {
        this.service.__set__('MemoryWatcher', {
          getInstance: this.sinon.stub().returns({
            liveness: true,
            start: () => {}
          })
        });
        const res = { status: this.sinon.stub(), json: this.sinon.stub() };
        this.service.init();

        await this.health.getMiddleware(this.health.ProbeTypes.Liveness)({}, res);
        expect(res.json.callCount).to.equal(1);
        expect(res.json.firstCall.args[0]).to.deep.equal({
          version: 'unknown',
          dependencies: {
            service: { status: 200, message: 'OK' },
            'memory-watcher': { status: 200, message: 'OK' }
          },
          state: 'Healthy'
        });
      });

      it('reports unhealthy when the "MemoryWatcher.liveness" property returns "false"', async () => {
        this.service.__set__('MemoryWatcher', {
          getInstance: this.sinon.stub().returns({
            liveness: false,
            start: () => {}
          })
        });
        const res = { status: this.sinon.stub(), json: this.sinon.stub() };
        this.service.init();

        await this.health.getMiddleware(this.health.ProbeTypes.Liveness)({}, res);
        expect(res.json.callCount).to.equal(1);
        expect(res.json.firstCall.args[0]).to.deep.equal({
          version: 'unknown',
          dependencies: {
            service: { status: 200, message: 'OK' },
            'memory-watcher': { status: 503, message: 'Liveness memory limit exceeded' }
          },
          state: 'Unhealthy'
        });
      });
    });

    describe('readiness check', () => {
      it('reports healthy when the "MemoryWatcher.readiness" property returns "true"', async () => {
        this.service.__set__('MemoryWatcher', {
          getInstance: this.sinon.stub().returns({
            readiness: true,
            start: () => {}
          })
        });
        const res = { status: this.sinon.stub(), json: this.sinon.stub() };
        this.service.init();

        await this.health.getMiddleware(this.health.ProbeTypes.Readiness)({}, res);
        expect(res.json.callCount).to.equal(1);
        expect(res.json.firstCall.args[0]).to.deep.equal({
          version: 'unknown',
          dependencies: {
            'memory-watcher': { status: 200, message: 'OK' },
            service: { status: 200, message: 'OK' }
          },
          state: 'Healthy'
        });
      });

      it('reports unhealthy when the "MemoryWatcher.liveness" property returns "false"', async () => {
        this.service.__set__('MemoryWatcher', {
          getInstance: this.sinon.stub().returns({
            readiness: false,
            start: () => {}
          })
        });
        const res = { status: this.sinon.stub(), json: this.sinon.stub() };
        this.service.init();

        await this.health.getMiddleware(this.health.ProbeTypes.Readiness)({}, res);
        expect(res.json.callCount).to.equal(1);
        expect(res.json.firstCall.args[0]).to.deep.equal({
          version: 'unknown',
          dependencies: {
            'memory-watcher': { status: 503, message: 'Readiness memory limit exceeded' },
            service: { status: 200, message: 'OK' }
          },
          state: 'Unhealthy'
        });
      });
    });
  });

  describe('Integration Test', () => {
    beforeEach(() => {
      this.sinon = sinon.createSandbox();
      this.service = require('../index');
    });

    afterEach(() => {
      this.sinon.restore();
    });

    it('Register Resources SIGTERM integration test', async () => {
      this.sinon.stub(this.service.process, 'exit');
      util.registerResources();
      const result = await util.runSigterm();
      expect(result).to.be.true;
    }).timeout(10000);
  });
});
