'use strict';

const { setTimeout } = require('timers/promises');
const { expect } = require('chai');
const sinon = require('sinon');
const rewire = require('rewire');
const MemoryWatcher = require('../lib/memory-watcher');

const mockMemoryGlobals = (rewiredModule, {
  totalMemory = 100,
  usedMemory = 50,
  existsSync,
  readFileSync
} = {}) => {
  const mocks = [
    rewiredModule.__set__('os', {
      totalmem: Number.isInteger(totalMemory) ? (() => totalMemory * 1024 * 1024) : totalMemory
    }),
    rewiredModule.__set__('process', {
      memoryUsage: {
        rss: Number.isInteger(usedMemory) ? (() => usedMemory * 1024 * 1024) : usedMemory
      },
      on: sinon.stub()
    }),
    rewiredModule.__set__('fs', {
      existsSync: existsSync || sinon.stub().returns(false),
      readFileSync: readFileSync || sinon.stub().returns('1')
    })
  ];
  return () => {
    mocks.forEach((revert) => revert());
  };
};

describe('MemoryWatcher', function () {
  describe('constructor', function () {
    it('returns a MemoryWatcher instance', function () {
      expect(new MemoryWatcher()).to.be.an.instanceOf(MemoryWatcher);
    });

    it('defaults "liveness" property to "false"', function () {
      const instance = new MemoryWatcher();
      expect(instance.liveness).to.be.false;
    });

    it('defaults "readiness" property to "false"', function () {
      const instance = new MemoryWatcher();
      expect(instance.readiness).to.be.false;
    });
  });

  describe('start', function () {
    let MockableMemoryWatcher;

    beforeEach(function () {
      MockableMemoryWatcher = rewire('../lib/memory-watcher');
    });

    it('logs config and memory statistics on startup', async function () {
      const logger = {
        info: sinon.stub(),
        silly: sinon.stub()
      };
      const instance = new MockableMemoryWatcher({ logger, healthyInterval: 1 });
      instance.start();
      await setTimeout(10);
      instance.stop();
      await setTimeout(10);
      expect(logger.info.callCount).to.equal(2);
      expect(logger.info.firstCall.args[0]).to.equal('exframe-service: starting memory-watcher with the following config');
      expect(logger.info.secondCall.args[0]).to.equal('exframe-service: current memory usage');
    });

    it('sets "liveness" to "true" and "readiness" to "true" when current memory usage is below all thresholds', async function () {
      const livenessThresholdOffset = 10;
      const readinessThresholdOffset = 20;
      const totalMemory = 50;
      const usedMemory = totalMemory - readinessThresholdOffset - 1;

      const revert = mockMemoryGlobals(MockableMemoryWatcher, { totalMemory, usedMemory });

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1,
        livenessThresholdOffset,
        readinessThresholdOffset
      });
      instance.start();
      await setTimeout(10);
      expect(instance.liveness).to.be.true;
      expect(instance.readiness).to.be.true;
      instance.stop();
      await setTimeout(10);
      revert();
    });

    it('sets "readiness" to "false" when current memory usage exceeds readiness threshold and sets "liveness" to "false" only after readinessThresholdInterval is reached', async function () {
      this.timeout(10000);
      const livenessThresholdOffset = 10;
      const readinessThresholdOffset = 20;
      const totalMemory = 50;
      const usedMemory = totalMemory - readinessThresholdOffset + 1;
      const revert = mockMemoryGlobals(MockableMemoryWatcher, { totalMemory, usedMemory });

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1,
        livenessThresholdOffset,
        readinessThresholdOffset,
        readinessThresholdInterval: 3000
      });

      instance.start();

      await setTimeout(10);

      expect(instance.readiness).to.be.false;
      expect(instance.liveness).to.be.true;

      await setTimeout(2000);

      expect(instance.readiness).to.be.false;
      expect(instance.liveness).to.be.true;

      await setTimeout(1000);
      expect(instance.readiness).to.be.false;
      expect(instance.liveness).to.be.false;
      // Clean up
      instance.stop();
      await setTimeout(10);
      revert();
    });

    it('sets "liveness" to "false" and "readiness" to "false" when current memory usage exceeds liveness threshold', async function () {
      const livenessThresholdOffset = 10;
      const readinessThresholdOffset = 20;
      const totalMemory = 50;
      const usedMemory = totalMemory - livenessThresholdOffset + 1;

      const revert = mockMemoryGlobals(MockableMemoryWatcher, { totalMemory, usedMemory });

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1,
        livenessThresholdOffset,
        readinessThresholdOffset
      });
      instance.start();
      await setTimeout(10);
      expect(instance.liveness).to.be.false;
      expect(instance.readiness).to.be.false;
      instance.stop();
      await setTimeout(10);
      revert();
    });

    it('resets "liveness" and "readiness" back to "true" when current memory usage drops below thresholds after exceeding them', async function () {
      const livenessThresholdOffset = 10;
      const readinessThresholdOffset = 20;
      const totalMemory = 50;
      const usedMemory = totalMemory - livenessThresholdOffset + 1;

      const revert = mockMemoryGlobals(MockableMemoryWatcher, { totalMemory, usedMemory });

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1,
        livenessThresholdOffset,
        readinessThresholdOffset
      });
      instance.start();
      await setTimeout(10);
      expect(instance.liveness).to.be.false;
      expect(instance.readiness).to.be.false;
      mockMemoryGlobals(MockableMemoryWatcher, {
        totalMemory,
        usedMemory: totalMemory - readinessThresholdOffset - 1
      });
      await setTimeout(10);
      expect(instance.liveness).to.be.true;
      expect(instance.readiness).to.be.true;
      instance.stop();
      await setTimeout(10);
      revert();
    });
  });

  describe('stop', function () {
    let MockableMemoryWatcher;

    beforeEach(function () {
      MockableMemoryWatcher = rewire('../lib/memory-watcher');
    });

    it('does not error if instance has not been started', async function () {
      const totalmemStub = sinon.stub();
      const revert = mockMemoryGlobals(MockableMemoryWatcher, { totalMemory: totalmemStub });

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1
      });
      instance.stop();
      expect(totalmemStub.callCount).to.equal(1); // 1 call from constructor
      revert();
    });

    it('prevents any subsequent memory checks', async function () {
      const totalmemStub = sinon.stub();
      const revert = mockMemoryGlobals(MockableMemoryWatcher, { totalMemory: totalmemStub });

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1
      });
      instance.start();
      await setTimeout(10);
      const memoryCheckCount = totalmemStub.callCount;
      instance.stop();
      await setTimeout(10);
      expect(totalmemStub.callCount).to.equal(memoryCheckCount);
      revert();
    });

    it('puts the instance in a state that allows it to be restarted', async function () {
      const logger = { info: sinon.stub(), silly: sinon.stub() };
      const instance = new MockableMemoryWatcher({
        logger,
        healthyInterval: 1,
        unhealthyInterval: 1
      });
      instance.start();
      await setTimeout(10);
      const logCallCount = logger.info.callCount;
      instance.stop();
      await setTimeout(10);
      expect(logger.info.callCount).to.equal(logCallCount);
      instance.start();
      await setTimeout(10);
      expect(logger.info.callCount).to.be.greaterThan(logCallCount);
      instance.stop();
    });
  });

  describe('getTotalMemory', function () {
    let MockableMemoryWatcher;

    beforeEach(function () {
      MockableMemoryWatcher = rewire('../lib/memory-watcher');
    });

    it('gets the total memory when limited with cgroup v1', function () {
      const existsSync = sinon.stub().onFirstCall().returns(true);
      const readFileSync = sinon.stub().onFirstCall().returns('128\n');
      const revert = mockMemoryGlobals(MockableMemoryWatcher, { existsSync, readFileSync });

      const totalMem = MockableMemoryWatcher.getTotalMemory();

      expect(totalMem).to.equal('128');
      revert();
    });

    it('gets the total memory when limited with cgroup v1, but is set to unlimited', function () {
      const existsSync = sinon.stub().onFirstCall().returns(true);
      const readFileSync = sinon.stub().onFirstCall().returns('-1\n');
      const totalMemory = sinon.stub().onFirstCall().returns('256');
      const revert = mockMemoryGlobals(MockableMemoryWatcher, { existsSync, readFileSync, totalMemory });

      const totalMem = MockableMemoryWatcher.getTotalMemory();

      expect(totalMem).to.equal('256');
      revert();
    });

    it('gets the total memory when limited with cgroup v2', function () {
      const existsSync = sinon.stub()
        .onFirstCall().returns(false)
        .onSecondCall().returns(true); // eslint-disable-line newline-per-chained-call
      const readFileSync = sinon.stub().onFirstCall().returns('128\n');
      const revert = mockMemoryGlobals(MockableMemoryWatcher, { existsSync, readFileSync });

      const totalMem = MockableMemoryWatcher.getTotalMemory();

      expect(totalMem).to.equal('128');
      revert();
    });

    it('gets the total memory when limited with cgroup v2, but is set to unlimited', function () {
      const existsSync = sinon.stub()
        .onFirstCall().returns(false)
        .onSecondCall().returns(true); // eslint-disable-line newline-per-chained-call
      const readFileSync = sinon.stub().onFirstCall().returns('max\n');
      const totalMemory = sinon.stub().onFirstCall().returns('256');
      const revert = mockMemoryGlobals(MockableMemoryWatcher, { existsSync, readFileSync, totalMemory });

      const totalMem = MockableMemoryWatcher.getTotalMemory();

      expect(totalMem).to.equal('256');
      revert();
    });

    it('gets the total memory when not limited by cgroups', function () {
      const existsSync = sinon.stub()
        .onFirstCall().returns(false)
        .onSecondCall().returns(false); // eslint-disable-line newline-per-chained-call
      const totalMemory = sinon.stub().onFirstCall().returns('512');
      const revert = mockMemoryGlobals(MockableMemoryWatcher, { existsSync, totalMemory });

      const totalMem = MockableMemoryWatcher.getTotalMemory();

      expect(totalMem).to.equal('512');
      revert();
    });
  });
});
